/*
 * Copyright (c) 2013, 2018 lvqcl
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "stdafx.h"
#define USE_LPC

#ifdef USE_LPC
#include "lpc/lpc.h"
#endif

static const double anti_denormal = 0.88817841970012523233890533447266e-15; //2e-50

class deemph_postprocessor_instance : public decode_postprocessor_instance
{
	struct fmem
	{
		double i1, o1, i2, o2;
	};

public:
	deemph_postprocessor_instance()
	{
		m_prev = false;
		m_buf = NULL;
		m_channels = m_channel_config = m_samplerate = 0;
	}

	~deemph_postprocessor_instance()
	{
		delete[] m_buf; m_buf = NULL;
		m_channels = m_channel_config = m_samplerate = 0;
	}

	virtual bool run(dsp_chunk_list& p_chunk_list, t_uint32 p_flags, abort_callback& p_abort)
	{
		t_size chunk_count = p_chunk_list.get_count();
		unsigned ret = 0;

		for(t_size i = 0; i < chunk_count; i++)
		{
			audio_chunk* chunk = p_chunk_list.get_item(i);
			ret |= do_chunk(chunk, p_abort);
		}
		return (ret != 0);
	}

	virtual bool get_dynamic_info(file_info& p_out)
	{
		return false;
	}

	virtual void flush()
	{
		if (m_buf) memset(m_buf, 0, sizeof(fmem)*m_channels);
	}

	virtual double get_buffer_ahead()
	{
		return 0;
	}

	unsigned do_chunk(audio_chunk* chunk, abort_callback&)
	{
		const unsigned sr = chunk->get_sample_rate();
		if (sr != 44100 && sr != 48000 && sr != 88200 && sr != 96000 && sr != 176400 && sr != 192000) { m_prev = false; return 0; }

		const unsigned channels = chunk->get_channels();
		const unsigned channel_config = chunk->get_channel_config();
		const t_size len = chunk->get_sample_count();
		audio_sample* const data0 = chunk->get_data();

		if (channels != m_channels || channel_config != m_channel_config || sr != m_samplerate || m_prev != true)
		{
			if (channels != m_channels)
			{
				delete[] m_buf;
				m_buf = new fmem[channels];
			}
			m_channels = channels; m_channel_config = channel_config; m_samplerate = sr; m_prev = true;
			deemph_postprocessor_instance::flush();

			switch(m_samplerate)
			{
			case 44100:
				init_coef(5277, 0.4850,  -9.465); break;
			case 48000:
				init_coef(5356, 0.4858,  -9.605); break;
			case 88200:
				init_coef(5635, 0.4970, -10.155); break;
			case 96000:
				init_coef(5660, 0.4975, -10.200); break;
			case 176400:
				init_coef(5764, 0.4993, -10.3745); break;
			case 192000:
				init_coef(5772, 0.4995, -10.388); break;
			default: // we cannot be here but ...
				m_prev = false; return 0;
			}
#ifdef USE_LPC
			if (len > 2*LPC_ORDER)
			{
				t_size siglen = min(len, 128);
				audio_sample* lpcbuf = new audio_sample[(PRE + siglen)*m_channels];
				audio_sample* xbuf = lpcbuf + PRE*m_channels;
				memmove(xbuf, data0, siglen*m_channels*sizeof(*data0));
				
				lpc_extrapolate_bkwd(xbuf, siglen, siglen, m_channels, LPC_ORDER, PRE);
				
				for(unsigned c=0; c<m_channels; c++)
				{
					audio_sample* data = lpcbuf + c;
					fmem& m = m_buf[c];
					for(t_size curr=0; curr < PRE; curr++)
					{
						double i0 = *data;
						i0 += anti_denormal; i0 -= anti_denormal;
						double o0 = i0*b0 + m.i1*b1 + m.i2*b2 - m.o1*a1 - m.o2*a2;
						o0 += anti_denormal; o0 -= anti_denormal;
						m.i2 = m.i1, m.i1 = i0;
						m.o2 = m.o1, m.o1 = o0;
						*data = (audio_sample) o0;
						data += m_channels;
					}
				}
				delete[] lpcbuf;
			}
#endif
		}

		for(unsigned c=0; c<m_channels; c++)
		{
			fmem& m = m_buf[c];
			audio_sample* data = data0 + c;
			for(t_size curr=0; curr < len; curr++)
			{
				double i0 = *data;
				// i0 += anti_denormal; i0 -= anti_denormal; // we assume that the input is always denormals-free

				double o0 = i0*b0 + m.i1*b1 + m.i2*b2 - m.o1*a1 - m.o2*a2;
				o0 += anti_denormal; o0 -= anti_denormal;

				m.i2 = m.i1, m.i1 = i0;
				m.o2 = m.o1, m.o1 = o0;

				*data = (audio_sample) o0;
				data += m_channels;
			}
		}
		return 1;
	}
private:
	bool m_prev;
	fmem* m_buf;
	unsigned m_channels, m_channel_config, m_samplerate;

#ifdef USE_LPC
	static const int LPC_ORDER = 32;
	static const int PRE = 1024;
#endif

	double b0, b1, b2, a1, a2;

	void init_coef(double fc, double slope, double gain)
	{
		const double w0 = 2 * M_PI * fc / m_samplerate;
		const double A  = exp(gain / 40 * M_LN10);
		const double alpha = sin(w0)/2 * sqrt((A + 1/A)*(1/slope - 1) + 2);

		const double cs = cos(w0);
		const double v = 2*sqrt(A)*alpha;
		double a0;

		b0 =    A*( (A+1) + (A-1)*cs + v );
		b1 = -2*A*( (A-1) + (A+1)*cs     );
		b2 =    A*( (A+1) + (A-1)*cs - v );
		a0 =        (A+1) - (A-1)*cs + v;
		a1 =    2*( (A-1) - (A+1)*cs     );
		a2 =        (A+1) - (A-1)*cs - v;

		b2 /= a0;
		b1 /= a0;
		b0 /= a0;
		a2 /= a0;
		a1 /= a0;
		a0 = 1.0;
	}
};


class deemph_postprocessor_entry : public decode_postprocessor_entry
{
public:
	virtual bool instantiate(const file_info& info, decode_postprocessor_instance::ptr& out)
	{
		int sr = (int)info.info_get_int("samplerate");
		if (sr != 44100 && sr != 48000 && sr != 88200 && sr != 96000 && sr != 176400 && sr != 192000) return false;

		const char* enabled = info.meta_get("pre_emphasis", 0);
		if (enabled == NULL) enabled = info.meta_get("pre-emphasis", 0);
		if (enabled == NULL) return false;

		if (pfc::stricmp_ascii(enabled, "1") == 0 || pfc::stricmp_ascii(enabled, "on") == 0 || pfc::stricmp_ascii(enabled, "yes") == 0)
		{
			out = new service_impl_t<deemph_postprocessor_instance>;
			return true;
		}
		return false;
	}
};

static service_factory_single_t<deemph_postprocessor_entry> g_deemph_postprocessor_entry;

// Declaration of my component's version information
DECLARE_COMPONENT_VERSION("De-emphasis postprocessor", "0.1.1", "Copyright(C) 2013, 2018 lvqcl\n"
"Usage: create tag PRE_EMPHASIS or PRE-EMPHASIS with value \"1\", \"on\" or \"yes\"\n"
);

// This will prevent users from renaming the component around (important for proper troubleshooter behaviors) or loading multiple instances of it.
VALIDATE_COMPONENT_FILENAME("foo_deemph.dll");
